/** 
 * "Hello file" -sovellus. Tulostaa tiedostoon ennalta arvattavan
 * tekstin.
 */
#define _POSIX_C_SOURCE 200809L
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

/** 
 * Yrittää luoda uuden tiedoston kirjoittamista varten; aliohjelma
 * palauttaa osoittimen tietovirtaolioon tai NULL, jos tiedosto oli jo
 * olemassa tai sitä ei voinut jostain syystä luoda.
 *
 * Kaiketi aika turvallinen tapa luoda tiedosto kirjoittamista varten:
 * soveltuisi tiedoston käyttämiseen lukituksena, koska open()
 * tapahtuu tässä atomisesti POSIXin määräämänä
 * käyttöjärjestelmäominaisuutena. Vain yksi prosessi/säie saa
 * tiedoston auki, ja muut yhdenaikaiset "kilpajuoksijat"
 * epäonnistuvat standardin mukaisesti.
 *
 * Helpommallakin saisi tehtyä tietovirran helloworldia varten
 * (fopen()), mutta tässä saa samalla demonstroitua
 * tiedostodeskriptoria, joka on käyttöjärjestelmän matalimman tason
 * rajapinta tiedoston käsittelyyn. Se on vaan kokonaisluku (jollainen
 * on nähty itse asiassa kurssin alkupuolen helloasm.s -esimerkin
 * syscallin parametrina).
 *
 * Käyttely on helpompaa C-kirjaston tarjoamalla kuorrutuksella (FILE
 * -objekti), joten täällä luodaan deskriptorin ympärille
 * "tietovirtaolio", joka palautetaan. Kutsuja on vastuussa tiedoston
 * sulkemisesta. Sulkeminen onnistuu kuorruteolion kautta.
 *
 */
FILE *luo_kirjoittamista_varten(const char *polku){
    FILE *res;
    int d = open(polku, O_CREAT | O_EXCL | O_WRONLY, S_IWUSR);
    if (d < 0) {
        if (errno == EEXIST) {
            fprintf(stderr, "Tiedosto '%s' on jo olemassa. En ylikirjoita!\n", polku);
        } else {
            perror("Ongelma tiedoston luonnissa");
        }
        return NULL;
    } else {
        res = fdopen(d, "w"); /* Deskriptorin kuorrutus C:n tietovirraksi.*/
        if (res == NULL){
            perror("Ongelma tiedoston paketoinnissa virraksi");
        } else {
            return res;
        }
    }    
    return NULL;
}

int main(int argc, char *argv[]){
    FILE *tied;
    if (argc < 2){
        fprintf(stderr, "Ilmoita tiedosto, johon kirjoitan.\n");
        exit(1);
    }
    tied = luo_kirjoittamista_varten(argv[1]);
    
    if (tied == NULL) {
        fprintf(stderr, "Tiedostoa ei saatu luotua.\n");
        exit(1);
    }
    
    fprintf(tied,
            "Hei, maailma! ... \n"
            "nyt POSIXin C-rajapinnan kautta avatussa tiedostossa! \n");
    
    if (fclose(tied) != 0){
        perror("Ongelma tiedoston sulkemisessa");
    }
    return 0;
}
